#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <sys/types.h>
#include <sys/bsd_types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "mgr.h"
/*#include "error.h"*/
#include "confmgr.h"
#include "confapi.h"

#define MSGLEN 512
#define size 64    /* max # of sessions and applications */
#define INT 0
#define CHAR 1

static void execute_child(void);
static int open_socket(void);
static void process_request(int sockfd);
static uint register_app(char *mesg);
static void unpack(char *mesg, appList *app);
static sessionList *findSession(void *param, int paramType);
static void addSession(sessionList *session, char *cname);
static uint addApp(sessionList *thisSession, appList *thisApp);
static appList *getApp(uint handle);
static appList *getPrevApp(uint handle);
static uint getHandle(appList *thisApp, sessionList *thisSession);

static  void *handles[size][size];    /* array for handles */
static sessionList *sessions = NULL;

void start_conf_mgr(void)
{
  switch(fork()) {
  case -1 : fprintf(stdout, " fork error in start_conf_mgr\n");
            exit(1);
  case 0  : execute_child();  /* conference manager */
  default : exit(0);          /* parent terminate */
  }
}

/* runs as a daemon process */
static void execute_child(void)
{
  int sockfd;

  sockfd = open_socket();
  if (sockfd == -1)
    err_dump("couldn't create socket for conference manager");

  process_request(sockfd);
}

static int open_socket(void)
{
  int                sockfd;
  struct sockaddr_in conf_mgr_addr;
    
  /* open a UDP socket */
  
  if ( (sockfd = socket(PF_INET, SOCK_DGRAM, 0) ) < 0)
    err_dump("Conference Manager : can't open a datagram socket");
  
  fprintf(stdout, "Conference Manager is coming up ....\n");
  
  /* Bind our local address so that the client applications can */
  /* communicate with us */
  
  bzero((char *) &conf_mgr_addr, sizeof(conf_mgr_addr));
  conf_mgr_addr.sin_family      = PF_INET;
  conf_mgr_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  conf_mgr_addr.sin_port        = htons(CONF_MGR_PORT);
  
  if (bind(sockfd, (struct sockaddr *) &conf_mgr_addr,
	   sizeof(conf_mgr_addr)) <0)
    err_dump("Conference Manager : can't bind local address");
  
  fprintf(stdout, "Conference Manager is up and ready ...\n");

  return sockfd;
}

static void process_request(int sockfd)
{
  int n, status;
  uint HANDLE;
  char mesg[MSGLEN+1];  /* buffer to store incoming message */

  struct sockaddr_in app_addr;

  while (1) {
    n = recvfrom(sockfd, mesg, MSGLEN, 0, &app_addr,
		 &sizeof(app_addr)); 
    if (n < 0)
      err_dump("process_request: recvfrom error");
        
    switch (mesg[2]){
    case CONFRegisterApp   : HANDLE = register_app(mesg);
                            /* send back the handle thru' sendto */
                            if (sendto(sockfd, (uint *)&HANDLE, sizeof(uint),
				       0, (struct sockaddr
					   *)&app_addr,
				       sizeof(app_addr)) == -1)
			      err_dump("can't send back handle");
      break;
    case CONFDeregisterApp : status = deregister_app(mesg);
                           /* send back the status thru' sendto */
                             if (sendto(sockfd, (int *)&status, sizeof(int),
					0, (struct sockaddr *)
					&app_addr, sizeof(app_addr))
				 == -1)
                                 err_dump("can't send back deregister
status"); 
      break;
    default  : fprintf(stdout, "process request : message not
                       recognized"); 
    }
  }
}


static uint register_app(char *mesg)
{
  uint HANDLE;    /* 4 byte handle for an application */
  sessionList *session;
 
  /* allocate memory on heap */
  applist *app = (appList *) malloc((size_t) sizeof(applist));
  if (!app)
    err_dump("register_app:memory allocation failure");

  unpack(mesg, app);
  session = findSession(app->CNAME, CHAR);
  if (!session) {
    /* allocate memory on heap */
    session = (sessionList *) malloc((size_t) sizeof(sessionList));
    if (!session)
      err_dump("register_app: memory allocation failure");
    else    /*add session with CNAME */
      addSession(session, app -> CNAME);
  }
  HANDLE = addApp(session, app); /* add application to a session */
  
  return HANDLE;
  
}  

static int deregister_app(mesg)
{
  uint handle;
  int status, result;
  appList *Deregister_app, *previous_app;
  sessionList *session;
  
  handle = (uint) atoi(mesg+3);
  if (Deregister_app = getApp(handle)) {
    previous_app = getPrevApp(handle);
    if (previous_app == Deregister_app) { /* application is first on
					     the list*/
      session = findSession(Deregister_app -> CNAME, CHAR);
      session -> head = Deregister -> next;
      Deregister_app -> next = NULL;
      if (session -> head == NULL) /*only one app. in the session */
	result = deregister_session(session, handle);
      free(Deregister_app);
      status = DEREGISTERED;
    }
    else {
      previous_app -> next = Deregister_app -> next;
      Deregister_app -> next = NULL;
      free (Deregister_app);
      status = DEREGISTERED;
    }
  }
  else
     status = INVALID_APPLICATION;

  return status;
}

static appList *getPrevApp(uint handle)
{
  int col;
  int row;

  col = (int) handle & 65535; 
  row = (int) handle >> ((sizeof(uint)/2) * 8);
  row--;
  if (row == -1) /* first application in the session */
    return (appList *) handles[row+1][col];
  else 
     return (appList *) handles[row][col];
}

static int deregister_session(sessionList *session, uint handle)
{
  sessionList *previous_session;

  previous_session = getPrevSession(session, handle);
  if (previous_session == session) { /*session is first on list */
    sessions = session -> next;
    session -> next = NULL;
    free(session);
  }
  else {                 
    previous_session -> next = session -> next;
    session -> next = NULL;
    free(session);
  }
  return 1;
}

static sessionList *getPrevSession(sessionList *session, uint handle)
{
  int sessionID;
 
  sessionID = (int) handle & 65535;   /* get current session ID */
  sessionID--;                        /* and then the previous one */
  if (sessionID == -1)     /* session is first on session list */
    return session;
  else {
    previous_session = findSession(&sessionID, INT);
    return previous_session;       
  }

}

/*
static sessionList *getSession(int sessionID)
{
  sessionList *scan;

  scan = sessions;
  
  while(scan){
    if (scan -> ID == sessionID)
      break;
    scan = scan -> next;
  }

  return scan;
}
*/

static void unpack(char *mesg, appList *app)
{
  app -> appType = (int) mesg[0];
  strcpy(app -> CNAME, mesg+3); 
  /* strncpy(app -> CNAME, mesg+3, *mesg[1] - 1);
     app -> CNAME[*mesg[1]] = '\0';
  */
}
  
  
  
/*static sessionList *findSession(char *cname) */

/*searches on the basis of either ID or CNAME */
static sessionList *findSession(void *param, int paramType)
{
  sessionList *scan;

  scan = sessions;
  
  if (paramType == INT)
    while(scan){
      if (scan -> ID == *(int *)param)
	break;
      scan = scan -> next;
    }
  else
    if (param == CHAR)
      while(scan){
	if (!strcmp(scan -> CNAME, (char *)param)) 
	  break;
	scan = scan -> next;
      }
  return scan;

}

/* add a session to the list of sessions */
static void addSession(sessionList *session, char *cname)
{
  static uint id = 0;

  if (!sessions)
    sessions = session; /* first session in the list */
  else {
    session -> next  = sessions;
    sessions = session;
  }
  
  /* fill in session structure */
  session -> ID = id++;
  strcpy(session -> CNAME, cname);
  session -> head = NULL;
}

static uint addApp(sessionList *thisSession, appList *thisApp)
{
  uint handle; 
  int temp;

  thisApp -> next = thisSession -> head;
  thisSession -> head = thisApp;

  handle = getHandle(thisApp, thisSession);

  return handle;

}

/* takes in a handle, returns pointer to application */
static appList *getApp(uint handle)
{
  int col;
  int row;

  col = (int) handle & 65535; 
  row = (int) handle >> ((sizeof(uint)/2) * 8);

  return (appList *) handles[row][col];     /* [col][row] */
}

static uint getHandle(appList *thisApp, sessionList *thisSession)
{
  uint handle;

  handle = thisApp -> ID << ((sizeof(uint)/2) *8);
  temp = thisSession -> ID & 65535;
  handle = handle | temp;

/*
  handle = thisSession -> ID << (sizeof(uint)/2);
  temp = thisApp -> ID & 255;
  handle = handle | temp;
*/
  /* add to list of handles */
  handles[thisApp -> ID][thisSession -> ID] = (void*) thisApp;

  return handle;
  
}

